# -*- coding: utf-8 -*-

import sys

from .error import  Invalid

import warnings
import logging, inspect
log = logging.getLogger(__name__)

_python3 = sys.version_info[0]>=3

class PASS:
    pass

class __MISSING__( str ):
    def __str__(self):
        return ''

    __unicode__ = __str__

MISSING = __MISSING__()

def _append_list(klass, key, data):
    key = "__%s__" % key
    setattr\
        ( klass
        , key
        , list(getattr(klass,key,[])) + list(data)
        )

def _merge_dict(klass, key, data):
    log.debug('merge dict %s in %s' % ( key, klass ))

    fields = dict(getattr\
        ( klass
        , '__%s__' % key
        , {}
        ))

    fields.update( data )

    setattr\
        ( klass
        , "__%s__" % key
        , fields
        )
   
    log.debug('%s.__%s__ = %s' % (  klass, key, getattr( klass, key ) ))

def _merge_fields(klass, key, fields):
    if (len(fields)%2 != 0) or (len(fields)<2):
        raise SyntaxError("Invalid number of fields supplied (%s). Use: %s(key, value, key, value, …)" % (len(fields),key))

    prev_fields = getattr\
        ( klass
        , '__%s__' % key
        , []
        )

    newfields = list(prev_fields)

    field_index = {}
    pos = 0
    for (name, value) in prev_fields:
        field_index[name] = pos
        pos+=1

    pos = 0
    for value in fields:
        if pos%2 != 0:
            if name in field_index:
                newfields[field_index[name]] = (name,value)
            else:
                newfields.append((name,value))
        else:
            name = value
        pos += 1

    setattr\
        ( klass
        , '__%s__' % key
        , newfields
        )


def _callback(klass):
    advice_data = klass.__dict__['__advice_data__']

    for key,(data, callback)  in advice_data.items():
        callback( klass, key, data)

    del klass.__advice_data__

    return klass

if not _python3:
    # no longer supported with py3
    from zope.interface.advice import addClassAdvisor

    def _advice(name, callback,  data, depth=3 ):
        frame = sys._getframe(depth-1)
        locals = frame.f_locals

        if (locals is frame.f_globals) or (
            ('__module__' not in locals) and sys.version_info[:3] > (2, 2, 0)):
            raise SyntaxError("%s can be used only from a class definition." % name)

        if not '__advice_data__' in locals:
            locals['__advice_data__'] = {}

        if name in locals['__advice_data__']:
            raise SyntaxError("%s can be used only once in a class definition." % name)


        if not locals['__advice_data__']:
            addClassAdvisor(_callback, depth=depth)

        locals['__advice_data__'][name] = (data, callback)


    def pre_validate(*validators):
        _advice('pre_validate', _append_list,  validators)

    def post_validate(*validators):
        _advice('post_validate', _append_list,  validators)

    def fieldset(*fields):
        _advice('fieldset', _merge_fields, fields )

def defaultErrorFormatter( context, error ):
    return error.message % error.extra


# TODO: don't set context.value to None when in fact "" is given
class Context( dict ):

    __value__ = MISSING
    __error__ = MISSING
    __result__ = MISSING

    parent = None
    root = None
    key = '/'

    isValidated = False
    isValidating = False

    taggedValidators = {}
    indexKeyRelation = {}
    numValues = 0

    def __init__(self, validator=None, value=MISSING, key='/', parent=None):
        if parent is not None:
            self.parent = parent
            self.root = parent.root
            self.key = key

            sep = self.root is not parent and '.' or ''

            self['path'] = '%s%s%s' % (parent.path,sep,key)
        else:
            self.root = self
            self.errorFormatter = defaultErrorFormatter
            self['path'] = key

        self.validator = validator
        self.value = value

    @property
    def path(self):
        return self['path']

    @property
    def childs(self):
        warnings.warn("Context.childs is deprecated. Please context.children instead", DeprecationWarning, stacklevel=2)
        return self.children

    @property
    def children(self):
        children = self.get('children',None)
        if children is None:
            children = self[ 'children' ] = {}
        return children

    @property
    def errorlist(self):
        errorlist = self.get('errorlist',None)
        if errorlist is None:
            errorlist = self[ 'errorlist' ] = []
        return errorlist

    @property
    def updates(self):
        updates = self.get('updates',None)
        if updates is None:
            updates = self[ 'updates' ] = []
        return updates

    @property
    def value(self):
        return self.get('value',self.__value__)

    @value.setter
    def value( self, value):
        if value is self.value:
            return

        if self.root.isValidating:
            self['value'] = value
            self.root.updates.append( self.path )
            return

        if (value == '') or value is [] or value is {}:
            value = None

        self.__value__ = value
        self.clear()

    @property
    def result(self):
        return self.validate()

    @property
    def error(self):
        return self.__error__.__unicode__()

    @error.setter
    def error( self, error ):
        self.__error__ = error
        error.context = self

        message = error.validator.__messages__[error.key]

        if message is not None:
            extra = error.data['extra']
            value = error.value
            data = error.data

            data['message'] = message
            if hasattr(error,'realkey'):
                data['key'] = error.realkey

            extra['value.type'] = getattr(value, '__class__', None) is not None \
                and getattr(value.__class__,'__name__', False) or 'unknown'

            value_str = value
            decode_utf8 = False

            if not isinstance(value, str):
                if _python3 and isinstance( value, bytes ):
                    decode_utf8 = True
                else:
                    try:
                        value_str = str(value)
                    except:
                        value_str = ''

            elif not _python3:
                decode_utf8 = True

            if decode_utf8:
                try:
                    value_str = value.decode('utf-8')
                except UnicodeDecodeError:
                    value_str = ''

            extra['value'] = value_str

            cache = getattr( self, 'cache', None)
            if cache is not None:
                extra.update( cache )

            self['error'] = self.__error__.data

            if self.__error__.context.path not in self.root.errorlist:
                self.root.errorlist.append( self.__error__.context.path )

    @property
    def validator(self):
        if not hasattr(self, '__validator__'):
            return None
        return self.__validator__

    @validator.setter
    def validator(self,value):
        self.__validator__ = value
        self.clear()

    def setIndexFunc( self, func ):
        self.__result__ = MISSING
        self.__error__ = MISSING
        self.indexKeyRelation = {}

        if func is not None:
            self.indexFunc = func
        elif hasattr( self, 'indexFunc'):
            del self.indexFunc

    def getKeyByIndex( self, index ):
        key = self.indexKeyRelation.get( index, None )
        if key is not None:
            return key

        indexFunc = getattr(self,'indexFunc',None)
        if indexFunc:
            if not self.indexKeyRelation:
                self.numValues = len(self.children)

            self.indexKeyRelation[ index ] = indexFunc( index )
            return self.indexKeyRelation[ index ]
        else:
            raise SyntaxError('Context %s has no children supporting indexing' % self.path)

    def clear( self, force=False ):
        if not self.isValidated and not force:
            return

        dict.clear( self )

        if self.parent is not None and self.parent.path:
            self['path'] = '%s.%s' % (self.parent.path,self.key)
        else:
            self['path'] = self.key

        self.isValidated = False

        self.__result__ = MISSING
        self.__error__ = MISSING

    def validate( self ):
        if self.isValidated:
            if self.__error__ is not MISSING:
                raise self.__error__
            return self.__result__

        self.isValidating = True

        if self.parent is not None:

            if not self.parent.isValidated and not self.parent.isValidating:
                self.parent.validate()

        if self.validator is None:
            raise AttributeError("No validator set for context '%s'" % self.path )

        try:
            result = self.validator.validate( self, self.__value__)

        except Invalid as e:
            self.error = e
            raise e
        else:
            if result is not PASS:
                self.__result__ = result
            else:
                self.__result__ = self.__value__

            return self.__result__
        finally:
            self.isValidated = True
            self.isValidating = False

    def __call__( self, path ):
        if path.__class__ is int:
            if path < 0:
                path = self.numValues+path

            return self( self.getKeyByIndex( path ) )
        elif not path:
            raise SyntaxError('Path cannot be empty')

        path = path.split('.',1)

        try:
            child = self.children[path[0]]
        except KeyError:
            child = self.children[path[0]] = Context( key=path[0], parent=self )

        if len(path) == 1:
            return child
        else:
            path=path[1]

        return child(path)


from .util import varargs2kwargs
# Some kind of 'clonable' object -
# we reinitialize child objects with inherited kwargs merged with new ones.
# This allows us to alter just a few specific parameters in child objects.
# without the need for implementors of validators to provide setters or too
# much specification for their attributes or how they are provided.
# * the setParameters function will be inspected, so that it will use
#   named parameters as kwargs, regardless if they are provided as *args
#   ( you cannot use *varargs in setParameters )
# * you can also define a setArguments function, which will be called before
#   setParameters, using the provided *varargs. keywords defined in
#   setParameters will not be moved from *args to **kwargs when setArguments
#   is defined. You can use it for attributes you only want to initialize
#   once. It also allows you to 'name' *varargs in the function definition.
# * __inherit__ specifies what attributes should be copied to child instances.
class Parameterized:
    __kwargs__ = {}
    __inherit__ = [ ]
    __isRoot__ = True

    __ignoreClassParameters__ = []

    def __init__( self, *args, **kwargs ):
        parent = kwargs.pop( '_parent', None )

        if not hasattr( self, 'setArguments') and args:
            func = getattr( self, 'setParameters', None)
            if func is not None:
                ( args, kwargs, shifted ) = varargs2kwargs( func, args, kwargs )

        if parent is not None:
            self.__isRoot__ = False
            newkwargs = dict(parent.__kwargs__ )
            newkwargs.update(kwargs)
            kwargs = newkwargs

            for key in self.__inherit__:
                setattr(self, key, getattr(parent, key))
        else:
            for key in self.__getParameterNames__():
                if hasattr(self.__class__,key)\
                and not key in self.__ignoreClassParameters__\
                and not key in kwargs:
                    kwargs[key] = getattr(self.__class__, key)

        if args or (parent is None):
            if hasattr( self, 'setArguments' ):
                self.setArguments( *args )
            elif args:
                raise SyntaxError('%s takes no further arguments' % self.__class__.__name__)

        if hasattr( self, 'setParameters' ):
            try:
                self.setParameters( **kwargs )
            except TypeError as e:
                raise TypeError(self.__class__.__name__+': '+e[0])

        elif kwargs:
            raise SyntaxError('%s takes no parameters' % self.__class__.__name__)

        self.__kwargs__ = kwargs

    def __call__( self, *args, **kwargs):
        kwargs['_parent'] = self
        return self.__class__( *args, **kwargs )


    @classmethod
    def __getParameterNames__( cls ):
        if not hasattr( cls, '__parameterNames__'):
            if not hasattr( cls, 'setParameters'):
                names = ()
            else:
                spec = inspect.getargspec( cls.setParameters )
                if spec.varargs:
                    raise SyntaxError('Cannot use *varargs in setParameters, please use %s.setArguments' % cls.__name__)
                names = spec.args[1:]
            setattr\
                ( cls,'__parameterNames__'
                , names
                )
        return cls.__parameterNames__

def inherit( *members ):
    def decorate( klass ):
        klassInherits = list(getattr(klass,'__inherit__', [] ))
        klassInherits += members

        setattr( klass, '__inherit__', klassInherits )
        return klass

    return decorate




from .core import ValidatorBase
from ..lib import Invalid, MISSING

class CacheBase( ValidatorBase ):

    def getCache( self, context, isGlobal=False ):
        if isGlobal:
            context = context.root

        cache = getattr( context, 'cache', None)

        if cache is None:
            cache = context.cache = {}

        return cache

class Set( CacheBase ):

    def __init__( self, key, value=MISSING ):
        self.key = key
        self.value = value

    def validate( self, context, value ):

        cache = self.getCache( context )
        val = value
        if self.value is not MISSING:
            val = self.value
        cache[self.key] = val

        return value

class Get( CacheBase ):
    def __init__( self, key ):
        self.key = key

    def validate( self, context, value ):
        cache = self.getCache( context )
        result = cache.get(self.key, MISSING)
        if result is MISSING:
            raise Invalid( value, self )

        return result





from ..lib import PASS, MISSING
from ..error import Invalid

from .core import Validator, ValidatorBase, messages

from copy import copy

import logging, sys

_python3 = sys.version_info[0]>=3

log = logging.getLogger(__name__)


@messages\
    ( fail='This field must be left out'
    )
class Missing( Validator ):

    def setParameters(self, default=PASS ):
        self.default = default

    def on_value( self, context, value ):
        raise Invalid( value, self )

    def on_missing( self, context ):
        if self.default is PASS:
            return MISSING
        return copy(self.default)
@messages\
    ( fail='This field must be blank'
    )
class Blank( Validator ):

    def setParameters( self, default = PASS ):
        self.default = default
        self.check_container = \
            isinstance( default, dict )\
            or  isinstance( default, list )\
            or  isinstance( default, tuple )

    def on_value( self, context, value ):
        if self.check_container \
        and not isinstance( value, str)\
        and ( isinstance( value, dict ) or isinstance( value, list ) or isinstance( value, tuple) ):
            n = MISSING
            if len(value) > 0:
                if isinstance( value, dict):
                    for (key, val) in value.items():
                        if val not in [ MISSING, None, '']:
                            n = value
                            break

                elif isinstance( value, list) or isinstance( value, tuple ):
                    for val in value:
                        if val not in [ MISSING, None, '']:
                            n = value
                            break
            if n is MISSING:
                return copy(self.default)

        raise Invalid( value, self )

    def on_blank( self, context, value ):
        if self.default is PASS:
            return value
        return copy(self.default)

@messages\
    ( fail='This field must be empty (missing or blank)'
    )
class Empty( Blank, Missing ):
    pass


@messages\
    ( fail='Value must match %(criterion)s'
    )
class Match( Validator ):

    RAW         = 'Match_RAW'
    REGEX       = 'Match_REGEX'
    VALIDATOR   = 'Match_VALIDATOR'

    def setParameters(self, criterion, ignoreCase=False):
        if not isinstance( criterion, ValidatorBase ):
            if hasattr(getattr( criterion, 'match', None ),'__call__'):
                self.type = Match.REGEX
            else:
                self.type = Match.RAW
        else:
            self.type = Match.VALIDATOR

        self.ignoreCase = ignoreCase
        self.criterion = criterion

    def appendSubValidators( self, subValidators ):
        if self.type == Match.VALIDATOR:
            self.criterion.appendSubValidators( subValidators )
            subValidators.append( self.criterion )

    def on_value(self, context, value ):
        if self.type is Match.REGEX:
            if _python3 and isinstance( value, bytes):
                value = value.decode('utf8')
            if not self.criterion.match(value):
                raise Invalid( value, self, matchType=self.type, criterion=self.criterion.pattern)
            return value
        elif self.type is Match.VALIDATOR:
            try:
                compare = self.criterion.validate( context, value )
            except Invalid as e:
                raise Invalid( value, self, matchType=self.type, criterion=e )
        else:
            compare = self.criterion

        val = value
        if self.ignoreCase:
            compare = str(compare).lower()
            val = str(value).lower()

        if val != compare:
            raise Invalid( value, self, matchType=self.type, criterion=compare )

        return value

    def on_missing(self, context ):
        if self.type is Match.VALIDATOR:
            return self.on_value( context, MISSING )
        return Validator.on_missing( self, context )

    def on_blank(self, context, value ):
        if self.type is Match.VALIDATOR:
            return self.on_value( context, value )
        return Validator.on_blank( self, context, value )


@messages\
    ( type="Can not get len from values of type %(value.type)s"
    , min="Value must have at least %(min)i elements/characters (has %(len)s)"
    , max="Value cannot have more than least %(max)i elements/characters (has %(len)s)"
    )
class Len( Validator ):

    def setParameters(self, min=1, max=None, returnLen=False):
        self.min = min
        self.max = max
        self.returnLen = returnLen

    def on_value(self, context, value):
        try:
            result = len(value)
        except Exception:
            raise Invalid( value, self, 'type' )

        if result<self.min:
            raise Invalid( value, self, 'min', min=self.min, max=self.max, len=result)
        if self.max is not None and result > self.max:
            raise Invalid( value, self, 'max', min=self.min, max=self.max, len=result)

        if self.returnLen:
            return result
        else:
            return value

@messages\
    ( fail="Value must be one of %(criteria)s"
    )
class In( Validator ):

    def setParameters( self, criteria ):
        self.criteria = criteria

    def on_value(self, context, value):
        if not value in self.criteria:
            raise Invalid( value, self, criteria=self.criteria )

        return value


@messages\
    ( fail="Value must lower or equal to %(max)s"
    )
class Max( Validator ):

    def setParameters( self, max ):
        self.max = max

    def on_value(self, context, value):
        if value > self.max:
            raise Invalid( value, self, max=self.max )

        return value



@messages\
    ( fail="Value must greater or equal to %(min)s"
    )
class Min( Validator ):

    def setParameters( self, min ):
        self.min = min

    def on_value(self, context, value):
        if value < self.min:
            raise Invalid( value, self, min=self.min )

        return value





# -*- coding: utf-8 -*-

from ..lib import PASS, MISSING, inherit
from ..error import Invalid

from .core import ValidatorBase, Validator, messages
from .check import Match


import logging
log = logging.getLogger(__name__)

@inherit\
    ( 'validators'
    , 'keyIndexRelation'
    , 'index'
    )
@messages\
    ( fail=None
    , extraFields='No extra fields allowed (extra fields: %(extraFields)s)'
    , type='Invalid input type (%(value.type)s) - must be dict, list or tuple.'
    )
class Schema( Validator ):

    def setArguments( self, *_fieldset ):
        if not hasattr(self,'__fieldset__') and ((len(_fieldset)%2 != 0) or (len(_fieldset)<2)):
            raise SyntaxError("Invalid number of fields supplied (%s). Use: %s(key, value, key, value, …)" % (len(_fieldset),self.__class__.__name__))

        fieldpairs = []
        pos = 0
        name = None
        for value in _fieldset:
            if pos%2 != 0:
                fieldpairs.append((name,value))
            else:
                name = value
            pos += 1

        (self.validators,self.index,self.keyIndexRelation)\
            = self.getValidators( fieldpairs or self.__fieldset__ )

        if not self.validators:
            raise SyntaxError('No fieldset given')

    def setParameters\
        ( self
        , allowExtraFields=False
        , returnList=False
        , createContextChildren=True
        ):

        self.returnList = returnList

        self.createContextChildren = createContextChildren
        self.allowExtraFields = allowExtraFields


    def appendSubValidators( self, subValidators ):
        for validator in list(self.validators.values()):
            validator.appendSubValidators( subValidators )
            subValidators.append(validator)

    def on_value( self, context, value ):
        if self.createContextChildren:
            self.on_value = self._createContextChildren_on_value
        else:
            self.on_value = self._on_value

        return self.on_value( context, value )

    def _on_value( self, context, value ):
        isList = isinstance(value, list) or isinstance(value,tuple) or isinstance(value,set)
        if not isList and not isinstance( value, dict ):
            raise Invalid( value, self, 'type')

        extraFields = None
        if not self.allowExtraFields:
            if isList:
                extraFields = max( len(value), len(self.index) )
            else:
                extraFields = list(value.keys())

        if self.returnList:
            result = []
        else:
            result = {}

        numValues = len(value)

        for pos in range(len(self.index)):
            key = self.index[pos]
            if isList:
                if numValues>pos:
                    val = value[ pos ]
                    if not self.allowExtraFields:
                        extraFields-=1
                else:
                    val = MISSING
            else:
                val = value.get( key, MISSING)
                if not self.allowExtraFields and val is not MISSING:
                    try: extraFields.remove(key)
                    except: pass

            res = self.validators[ key ].validate( context, val )
            if self.returnList:
                result.append( res )
            else:
                result[ key ] = res

        if extraFields:
            raise Invalid( value, self, 'extraFields',extraFields=extraFields)

        return result


    def _createContextChildren_on_value( self, context, value ):
        isList = isinstance(value, list) or isinstance(value,tuple) or isinstance(value,set)

        if not isList and not isinstance( value, dict ):
            raise Invalid( value, self, 'type')

        extraFields = None
        if not self.allowExtraFields:
            if isList:
                extraFields = max( len(value), len(self.index) )
            else:
                extraFields = list(value.keys())

        errors = []


        if self.returnList:
            result = []
        else:
            result = {}

        len_value = len(value)
        len_index = len(self.index)

        # populate
        for pos in range(len_index):
            key = self.index[pos]
            childContext = context( key )
            try:
                childContext.validator = self.validators[ key ]
            except KeyError:
                raise SyntaxError("No validator set for %s" % childContext.path)

            if isList:
                if len_value<=pos:
                    childContext.__value__ = MISSING
                else:
                    childContext.__value__ = value[ pos ]
            else:
                childContext.__value__ = value.get( key, MISSING )

            if not self.allowExtraFields:
                if isList:
                    extraFields-=1
                else:
                    try: extraFields.remove(key)
                    except: pass

        if extraFields:
            raise Invalid( value, self, 'extraFields',extraFields=extraFields)

        context.setIndexFunc( lambda index: self.index[index] )

        # validate
        for key in self.index:

            try:
                res = context( key ).result
            except Invalid as e:
                errors.append( e.context.key )
            else:
                if self.returnList:
                    result.append( res )
                else:
                    result[ key ] = res


        if errors:
            raise Invalid( value, self, errors=errors )

        return result


    @classmethod
    def getValidators( cls, _fieldset  ):
        if not _fieldset:
            return None

        validators = {}
        keyIndexRelation = {}
        index = []

        pos = 0
        for (name,validator) in _fieldset:
            validators[name] = validator
            keyIndexRelation[ name ] = pos
            index.append( name )
            pos += 1

        return validators,index,keyIndexRelation

@messages\
    ( fail=None
    , numericKeys='Invalid keys, please use 0,1,2,... (keys: %(keys)s)'
    , type='Invalid input type (%(value.type)s) - must be dict, list, tuple or set'
    , listType='Invalid input type (must be list, tuple or set)'
    )
class ForEach( Validator ):

    def setParameters\
        ( self
        , criterion
        , numericKeys=True
        , returnList=True
        , createContextChildren=True ):

        if not isinstance( criterion, ValidatorBase ):
            criterion = Match( criterion )

        self.returnList = returnList
        self.numericKeys = numericKeys
        self.validator = criterion
        self.createContextChildren = createContextChildren

    def on_value( self, context, value ):
        if self.createContextChildren:
            self.on_value = self._createContextChildren_on_value
        else:
            self.on_value = self._on_value

        return self.on_value( context, value )

    def appendSubValidators( self, subValidators ):
        self.validator.appendSubValidators( subValidators )
        subValidators.append( self.validator )

    def _on_value( self, context, value ):
        if self.returnList:
            result = []
        else:
            result = {}

        isList = isinstance( value, list) or isinstance(value, tuple) or isinstance(value, set)
        if not isList:
            if not isinstance(value, dict ):
                raise Invalid( value, self,'type' )

        if isList or self.numericKeys:
            for pos in range( len( value ) ):
                if not isList:
                    val = value.get(str(pos),MISSING)
                    if val is MISSING:
                        raise Invalid( value, self, 'numericKeys', keys=list(value.keys()) )
                else:
                    val = value[pos]

                res = self.validator.validate( context, val )

                if self.returnList:
                    result.append( res )
                else:
                    result[pos] = res
        else:
            for (key, val) in value.items():

                res = self.validator.validate( context, val )

                if self.returnList:
                    result.append( res )
                else:
                    result[key] = res

        return result

    def _createContextChildren_on_value( self, context, value ):
        isList = isinstance( value, list) or isinstance(value, tuple) or isinstance(value, set)

        if not isList:
            if not isinstance(value, dict ):
                raise Invalid( value, self,'type' )

        if self.returnList:
            result = []
        else:
            result = {}
        errors = []

        # populate
        children = []
        if isList or self.numericKeys:
            context.setIndexFunc( lambda index: str(index) )

            for pos in range( len( value ) ):
                if not isList:
                    val = value.get(str(pos),MISSING)
                    if value.get(str(pos),MISSING) is MISSING:
                        context.setIndexFunc( None )
                        raise Invalid( value, self, 'numericKeys',keys=list(value.keys()))

                else:
                    val = value[ pos ]

                contextChild = context( str( pos ) )
                contextChild.validator = self.validator
                contextChild.__value__ = val
                children.append( contextChild )

        else:
            context.setIndexFunc( None )

            if self.returnList:
                raise Invalid( value, self, 'listType' )
            for (key,val) in value.items():
                contextChild = context( key )
                contextChild.validator = self.validator
                contextChild.__value__ = val
                children.append( contextChild )

        #validate
        for childContext in children:
            try:
                res = childContext.result
            except Invalid:
                errors.append( childContext.key )
            else:
                if self.returnList:
                    result.append( res )
                else:
                    result[ childContext.key ] =  res

        if errors:
            raise Invalid( value, self, errors=errors )

        return result


class FieldValidator( Validator ):

    def setParameters(self):
        raise SyntaxError( "FieldValidator cannot be used directly" )

    def getField( self, context, path):
        pathSplit = path.split('.')

        if pathSplit[0].startswith('/'):
            pathSplit[0] = pathSplit[0][1:]
            fieldcontext = context.root
        else:
            fieldcontext = context

            while fieldcontext.parent and pathSplit and pathSplit[0] is '':
                fieldcontext = fieldcontext.parent
                del pathSplit[0]

        for part in pathSplit:
            if part.startswith('(') and part.endswith(')'):
                part = int(part[1:-1])

            fieldcontext = fieldcontext( part )

        if fieldcontext is context:
            raise SyntaxError( "Cannot reference myself. Nice try, though :)" )

        return fieldcontext

@messages\
    ( noResult='Field %(path)s has no result'
    )
class Field( FieldValidator ):

    def setParameters(self, path, criterion=None, useResult=False, copy=False, writeToContext=False):
        self.path = path
        self.copy = copy
        self.useResult = useResult
        self.writeToContext = writeToContext

        if criterion is None:
            self.copy = True
        elif not isinstance( criterion, ValidatorBase):
            criterion = Match( criterion )

        self.validator  = criterion

    def appendSubValidators( self, subValidators ):
        self.validator.appendSubValidators( subValidators )
        subValidators.append( self.validator )

    def validate(self, context, value):
        fieldcontext = self.getField( context, self.path )

        if not self.useResult:
            result = fieldcontext.value

        else:
            try:
                result = fieldcontext.result
            except Invalid:
                result = PASS

        if self.validator is not None:
            if result is not PASS:
                result = self.validator.validate( fieldcontext, result )

        if self.writeToContext:
            fieldcontext.__result__ = result

        if self.copy:
            if result is PASS:
                return value

            return result

        return value



from .core import Validator, validator2parameter, messages

from ..error import Invalid

from datetime import date, datetime

import logging, sys
log = logging.getLogger(__name__)

_python3 = sys.version_info[0]>=3

class TypeValidator( Validator ):
    __ignoreClassParameters__ = 'convert'

    converter = None

    def setParameters( self, convert=False ):
        self._convert = convert

    @classmethod
    def convert( cls, *args, **kwargs ):
        if not cls.converter:
            kwargs['convert'] = True
            cls.converter = cls( *args, **kwargs )
        return cls.converter

@messages\
    ( type="Invalid type (%(value.type)s), must be a dictionary"
    , convert="Could not convert %s(type)s to dict"
    )
class Dict( TypeValidator ):

    def on_value(self, context, value):

        if not isinstance(value, dict):
            if  not self._convert:
                raise Invalid( value, self, 'type' )
            try:
                value = dict(value)
            except (ValueError,TypeError):
                raise Invalid( value, self,'convert')

        if len( value ) == 0:
            return self.on_blank( context, value )

        return value

@messages\
    ( type="Invalid type (%(value.type)s), must be a list"
    , convert="Could not convert %(value.type)s to list"
    )
class List( TypeValidator ):

    def on_value(self, context, value):
        if isinstance(value,set) or isinstance(value,tuple):
            value = list(value)

        if not isinstance(value, list):
            if not self._convert:
                raise Invalid( value, self,'type' )

            try:
                value = list(value)
            except (ValueError,TypeError):
                raise Invalid( value, self,'convert' )

        if len( value ) == 0:
            return self.on_blank( context, value )

        return value

@messages\
    ( type="Invalid type (%(value.type)s), must be a bool"
    , convert="Could not convert %(value.type)s to bool"
    )
class Boolean( TypeValidator):

    def setParameters( self, convert=False ):
        TypeValidator.setParameters( self, convert )

    def on_value(self, context, value):
        if not (isinstance( value, bool )):
            if not self._convert:
                raise Invalid( value, self, 'type' )
            else:
                try:
                    value = str(value).lower()
                except Exception as e:
                    raise Invalid( value, self, 'convert' )
                if value in ('1', 'true', 'yes', 'on'):
                    value = True
                elif value in ('0', 'false', 'no', 'off'):
                    value = False
                else:
                    raise Invalid( value, self, 'convert' )
        return value

@messages\
    ( type="Invalid type (%(value.type)s), must be a string"
    , convert="Could not convert %(value.type)s to string"
    )
class String( TypeValidator ):

    def setParameters( self, convert=False ):
        TypeValidator.setParameters( self, convert )

    def on_value_py2( self, context, value ):
        if not isinstance( value, basestring):
            if not self._convert:
                raise Invalid( value, self, 'type' )
            else:
                try:
                    value = value.__str__()
                except AttributeError:
                    raise Invalid( value, self, 'convert' )

        return value

    def on_value(self, context, value):
        if not isinstance( value, str):
            if not self._convert:
                raise Invalid( value, self, 'type' )
            else:
                try:
                    value = value.__str__()
                except AttributeError:
                    raise Invalid( value, self, 'convert' )


        return value

if not _python3:
    String.on_value = String.on_value_py2

@messages\
    ( type="Invalid type (%(value.type)s), must be a integer"
    , convert="Could not convert %(value.type)s to integer"
    )
class Integer( TypeValidator ):

    def on_value(self, context, value):
        if not isinstance( value, int ) and not isinstance( value, int):
            if not self._convert:
                raise Invalid( value, self,'type' )
            try:
                value = int(value)
            except (TypeError, ValueError):
                raise Invalid( value, self,'convert' )

        return value

@messages\
    ( type="Invalid type (%(value.type)s), must be a floating point number"
    , convert="Could not convert %(value.type)s to a floating point number"
    )
class Float( TypeValidator ):

    def on_value(self, context, value):
        if not isinstance(value,float):
            if not self._convert:
                raise Invalid( value, self,'type')
            try:
                value = float(value)
            except (TypeError, ValueError):
                raise Invalid( value, self,'convert' )

        return value

@messages\
    ( type="Invalid type (%(value.type)s), must be a date"
    , convert='Could not convert "%(value)s"(%(value.type)s) to a date'
    )
class Date( Validator ):

    def setParameters( self, formatter="%Y-%m-%d", convert=False ):
        self._convert = convert
        validator2parameter(self, 'formatter', formatter)

    def on_value(self, context, value ):

        if not isinstance( value, date):
            if not self._convert:
                raise Invalid( value, self, 'type' )
            else:
                try:
                    return datetime.strptime( value, context.params.formatter ).date()
                except ValueError:
                    raise Invalid( value, self, 'convert' )
            
        return value

    @classmethod
    def convert( cls, formatter ):
        return cls( formatter=formatter, convert=True )


@messages\
    ( type="Invalid type (%(value.type)s), must be a datetime"
    , convert='Could not convert "%(value)s"(%(value.type)s) to a datetime'
    )
class DateTime( Validator ):

    def setParameters( self, formatter="%Y-%m-%d", convert=False ):
        self._convert = convert
        validator2parameter(self, 'formatter', formatter)

    def on_value(self, context, value ):

        if not isinstance( value, datetime):
            if not self._convert:
                raise Invalid( value, self, 'type' )
            else:
                try:
                    return datetime.strptime( value, context.params.formatter )
                except ValueError:
                    raise Invalid( value, self, 'convert' )
            
        return value

    @classmethod
    def convert( cls, formatter ):
        return cls( formatter=formatter, convert=True )



from .validator import *
from .lib import MISSING, Context

import pkg_resources as __pkg_resources__
__pkg_resources__.declare_namespace(__name__)
__version__ = __pkg_resources__.require("kanone")[0].version



import inspect

def varargs2kwargs( function, args, kwargs, skipSelf=True ):
    names = getParameterNames( function, skipSelf )

    realargs = list(args)
    shifted = []

    for argpos in range(min(len( names ), len(args))):
        key = names[ argpos ]
        if key in kwargs:
            raise SyntaxError('multiple kw args: %s' % key)
        kwargs[ key ] = args[argpos]
        shifted.append( key )
        del realargs[0]

    return realargs,kwargs,shifted

def getArgSpec( function ):
    function = getattr( function, '__func__', function )
    spec = getattr( function, '__spec__', None)
    if spec is None:
        function.__spec__ = spec = inspect.getargspec( function )
    return spec

def getParameterNames( function, skipSelf=True ):
    function = getattr( function, '__func__', function )
    names = getattr( function, '__parameterNames__', None)
    if names is None:
        spec = getArgSpec( function )
        function.__parameterNames__ = names = skipSelf and spec.args[0] is 'self' and spec.args[1:] or spec.args
    return names




from ..lib import Invalid

from .core import Validator, ValidatorBase, messages

@messages\
    ( type = "Values of type %(value.type)s can not be lowered"
    )
class Lower( Validator ):


    def on_value(self, context, value):
        try:
            return value.lower()
        except Exception:
            raise Invalid( value, self, 'type' )

@messages\
    ( type = "Cannot format type %(value.type)s"
    )
class Format( Validator ):


    def setParameters( self, formatter, **parameters ):
        self.formatter = formatter
        self.parameters = parameters

    def on_value(self, context, value):
        parameters = dict(self.parameters)
        parameters['value'] = value

        if not isinstance( self.formatter, str ):
            raise Invalid( value, self, 'type' )
        try:
            return self.formatter % parameters
        except Exception:
            raise Invalid( value, self )
@messages\
    ( type = "Cannot update type %(value.type)s"
    )
class DictUpdate( Validator ):

    def setParameters( self, **parameters ):
        self.parameters = parameters

    def on_value(self, context, value ):
        toUpdate = {}
        for (key,param) in self.parameters.items():
            if isinstance(param,ValidatorBase):
                param = param.validate( context, value )
            toUpdate[key] = param
        try:
            value.update( toUpdate )
        except Exception:
            raise Invalid( value, self, 'type' )

        return value

@messages\
    ( type = "Can not eliminate white spaces in values of type %(value.type)s"
    )
class EliminateWhiteSpace( Validator ):

    def on_value( self, context, value):
        try:
            return ''.join(value.split())
        except AttributeError:
            raise Invalid( value, self, 'type' )

# TODO: don't set context.value to None when in fact "" is given, see lib
@messages\
    ( type = "Can not strip white spaces in values of type %(value.type)s"
    )
class Strip( Validator ):

    def on_value( self, context, value):
        try:
            return value.strip()
        except AttributeError:
            raise Invalid( value, self, 'type' )

    on_blank = on_value

@messages\
    ( type = "Can not split values of type %(value.type)s"
    )
class Split( Validator ):

    def setParameters(self, separator=None, limit=-1):
        self.separator = separator
        self.limit = limit

    def on_value( self, context, value):
        try:
            return value.split( self.separator, self.limit )
        except Exception:
            raise Invalid( value, self, 'type' )

@messages\
    ( type = "Can not join values of type %(value.type)s"
    )
class Join( Validator):

    def setParameters(self, separator=''):
        self.separator = separator

    def on_value( self, context, value):
        try:
            return self.separator.join( value )
        except Exception:
            raise Invalid( value, self, 'type' )

@messages\
    ( type = "Can not encode %(value.type)s to %(format)s"
    , fail = "%(value)s cannot be encoded to %(format)s"
    )
class Encode( Validator ):

    def setParameters( self, format ):
        self.format = format

    def on_value( self, context, value ):
        if not hasattr( value,'encode') or not hasattr( value.encode,'__call__' ):
            raise Invalid( value, self, 'type', format=self.format )

        try:
            value = value.encode( self.format )
        except ValueError:
            raise Invalid( value, self, format=self.format )

        return value

@messages\
    ( type = "Can not decode %(value.type)s to %(format)s"
    , fail = "%(value)s cannot be decoded to %(format)s"
    )
class Decode( Validator ):

    def setParameters( self, format ):
        self.format = format

    def on_value( self, context, value ):
        if not hasattr( value,'decode') or not hasattr( value.decode,'__call__' ):
            raise Invalid( value, self, 'type', format=self.format )

        try:
            value = value.decode( self.format )
        except ValueError:
            raise Invalid( value, self, format=self.format )

        return value

@messages\
    ( type = "Unsupported type for insertion (%(value.type)s)"
    , fail = "Can not insert %(what)s at %(where)i"
    )
class Insert( Validator ):

    def setParameters( self, what, where=0 ):
        self.what = what
        self.where = where

    def on_value( self, context, value ):
        if ( isinstance( value, str ) and isinstance(value,str) )\
        or ( isinstance( value, list ) and isinstance( value, list ) ):
            if self.where is 0:
                return self.what + value
            if self.where <0:
                where = self.where-1
            else:
                where = self.where
            return value[0:where] + self.what + value[where:None]

        raise Invalid( value, self,'type' )

class UpdateValue( ValidatorBase ):

    def validate( self, context, value ):
        context.value = value
        return value



from .core import ValidatorBase

class Print( ValidatorBase ):

    def __init__( self, formatter):
        self.formatter=formatter

    def validate( self, context, value):
        print(( self.formatter % {'value':value} ))
        return value
